/*
 * Copyright ©2009 tpoinsot
 * Les sources qui constituent ce projet "Soma" de même que la documentation
 * associée sont la propriété de leur auteur.
 * Je donne mon accord au site developpez.com pour l'utilisation de tout ou
 * partie des sources et de la documentation de ce projet dans les pages developpez.com
 *
 * Piece : liste de cubes formant une pièce de base, et toutes ses positions
 * possible par rotation autour de l'un de ses cubes.
 *
 * Matrices de rotation :
 *
 *         (   1        0        0     )
 * Rotx(a) (   0      cos(a)  -sin(a)  )
 *         (   0      sin(a)   cos(a)  )
 * 
 *         ( cos(b)     0      sin(b)  )
 * Roty(b) (   0        1        0     )
 *         ( -sin(b)    0      cos(b)  )
 *
 *         ( cos(c)  -sin(c)     0     )
 * Rotz(c) ( sin(c)   cos(c)     0     )
 *         (   0        0        1     )
 *
 * Ici, a, b et c sont des multiples de PI/2.
 */

package soma;

import java.util.ArrayList;

/**
 *
 * @author tpoinsot
 */
public class Piece {
    private int numéro;
    private ArrayList<Cube> formeBase = new ArrayList<Cube>();
    
    private ArrayList<Piece> positions = new ArrayList<Piece>();
    private int positionIndex;

    /* matrices de rotation :
     * ( a b c )
     * ( d e f )
     * ( g h i )
     * écrite { a, b, c, d, e, f, g, h, i }
     *
     * Pour obtenir les matrices ci-dessous :
     *  4 rotations de PI/2 autour de x
     * puis PI/2 autour de y et 4 rotations de PI/2 autour de x
     * puis PI/2 autour de z et 4 rotations de PI/2 autour de x
     * puis PI/2 autour de y et 4 rotations de PI/2 autour de x
     * puis PI/2 autour de z et 4 rotations de PI/2 autour de x
     * puis PI/2 autour de y et 4 rotations de PI/2 autour de x
     * 4x6 = 24 positions.
     */
    static final int [][] mat = {
        { 1, 0, 0, 0, 1, 0, 0, 0, 1 },
        { 1, 0, 0, 0, 0, -1, 0, 1, 0 },
        { 1, 0, 0, 0, -1, 0, 0, 0, -1 },
        { 1, 0, 0, 0, 0, 1, 0, -1, 0 },
        { 0, 0, 1, 0, 1, 0, -1, 0, 0 },
        { 0, 1, 0, 0, 0, -1, -1, 0, 0 },
        { 0, 0, -1, 0, -1, 0, -1, 0, 0 },
        { 0, -1, 0, 0, 0, 1, -1, 0, 0 },
        { 0, 0, 1, 1, 0, 0, 0, 1, 0 },
        { 0, 1, 0, 1, 0, 0, 0, 0, -1 },
        { 0, 0, -1, 1, 0, 0, 0, -1, 0 },
        { 0, -1, 0, 1, 0, 0, 0, 0, 1 },
        { -1, 0, 0, 0, 0, 1, 0, 1, 0 },
        { -1, 0, 0, 0, 1, 0, 0, 0, -1 },
        { -1, 0, 0, 0, 0, -1, 0, -1, 0 },
        { -1, 0, 0, 0, -1, 0, 0, 0, 1 },
        { 0, 1, 0, 0, 0, 1, 1, 0, 0 },
        { 0, 0, -1, 0, 1, 0, 1, 0, 0 },
        { 0, -1, 0, 0, 0, -1, 1, 0, 0 },
        { 0, 0, 1, 0, -1, 0, 1, 0, 0 },
        { 0, 1, 0, -1, 0, 0, 0, 0, 1 },
        { 0, 0, -1, -1, 0, 0, 0, 1, 0 },
        { 0, -1, 0, -1, 0, 0, 0, 0, -1 },
        { 0, 0, 1, -1, 0, 0, 0, -1, 0 }
    };

    public Piece()
    {
    }

    public Piece(int index)
    {
        this.numéro = index;
    }

    public void addCube(int x, int y, int z)
    {
        formeBase.add(new Cube(x, y, z));
    }

    public void addCube(Cube p)
    {
        formeBase.add(new Cube(p));
    }

    /**
     * Initialise les positions d'une pièce de base
     */
    public void préparer()
    {
        int nfb = formeBase.size();
        positions.clear();
        for (int p = 0; p < nfb; p++) {
            // Changement de repère : Premier point de p en (0,0,0)
            Cube qp = formeBase.get(p);

            // toutes les rotations
            for (int x = 0; x < 24; x++) {
                Piece pnp = new Piece(numéro);
                // Mettre en premier le point "origine", car lui ira toujours,
                // puisqu'on va le placer à la première case libre et vérifier
                // les autres cubes
                pnp.addCube(formeBase.get(p).moveNeg(qp).multMat(mat[x]));
                for (int i = 0; i < nfb; i++) {
                    if ( i != p )
                        pnp.addCube(formeBase.get(i).moveNeg(qp).multMat(mat[x]));
                }
                // Ne pas insérer de doublons
                boolean bDéjàLà = false; // a-t-on déjà cette position ?
                for (int j = 0; j < positions.size(); j++) {
                    if ( positions.get(j).isEqual(pnp) ) {
                        bDéjàLà = true;
                        break;
                    }
                }
                if ( bDéjàLà == false ) {
                    positions.add(pnp);
                }
            }
        }
    }

    public Piece positionFirst()
    {
        positionIndex = 0;
        return positions.get(positionIndex);
    }

    public Piece positionNext()
    {
        positionIndex++;
        return positionIndex < positions.size() ? positions.get(positionIndex) : null;
    }

    /**
     * Met en place la pièce (qui est une position de la pièce d'origine)
     * à la première case du tableau "casesLibres".
     * La pièce a été orientée (avec préparer()) et positionnée dans l'espace :
     * Il reste a faire une translation vers la case libre et vérifier que 
     * chaque cube de la pièce correspond à une case libre.
     * @param casesLibres
     * @return Si la pièce a trouvé sa place, une copie du tableau
     * casesLibres diminué de la pièce, et Null dans le cas contraire.
     */
    public ArrayList<Cube> placer(ArrayList<Cube> casesLibres)
    {
        int nCasesLibres = casesLibres.size();
        if ( nCasesLibres == 0 ) 
            return null;

        int nCubesBase = formeBase.size(); // nombre de cubes de la forme de base
        // nCubesBase > 0 sinon on s'est trompé dans la définition des pièces

        int i, j;
        Cube c0 = casesLibres.get(0); // déplacement par rapport à l'origine

        // casesOk va conserver les bonnes cases : si j'en obtiens autant que
        // de cubes dans la pièce alors c'est gagné
        Cube [] casesOk = new Cube[nCubesBase];
        int nCasesOk = 0;

        // On sait que le 1er cube est en c0, donc i et j commencent à 1
        // j'ajouterai c0 à casesOk après, si tout va bien
        for (i = 1; i < nCubesBase; i++) {
            // qi est une copie déplacée : les cubes de base n'ont pas bougé
            Cube qi = formeBase.get(i).move(c0);
            boolean trouvé = false;
            for (j = 1; j < nCasesLibres; j++) {
                Cube qj = casesLibres.get(j);
                if ( qi.positionIdentique(qj) ) {
                    casesOk[nCasesOk++] = qj;
                    trouvé = true;
                    break;
                }
            }
            // Si l'un des cubes de la pièce ne rentre pas, c'est fini et pas bon
            if ( trouvé == false )
                return null;
        }
        //-------------------------------------------------------------------
        // Si on arrive là, c'est que les nCubesBase cubes ont été trouvés
        // Donc il reste à marquer la place ...
        casesOk[nCasesOk++] = c0;
        for (j = 0; j < nCasesOk; j++) {
            casesOk[j].numeroPiece = numéro;
        }
        
        // ... et réduire de la liste des cases vides
        ArrayList<Cube> listeRetour = new ArrayList<Cube>(nCasesLibres - nCubesBase);
        for (i = 0; i < nCasesLibres; i++) {
            Cube qRet = casesLibres.get(i);
            if ( qRet.numeroPiece == 0 )
                listeRetour.add(qRet);
        }
        return listeRetour;
    }

    /**
     * Libère les cubes occupés par cette pièce => remet numeroPiece à 0
     * @param lp
     */
    public void libérer(ArrayList<Cube> lq)
    {
        int nj = lq.size();
        for (int j = 0; j < nj; j++)
        {
            Cube qj = lq.get(j);
            if ( qj.numeroPiece == numéro )
                qj.numeroPiece = 0;
        }
    }

    public boolean isEqual(Piece p)
    {
        // p.formeBase == formeBase ? quel que soit l'ordre des cubes dans les pièces
        if ( p.formeBase.size() != this.formeBase.size() )
            return false;

        int n = formeBase.size();
        for (int i = 0; i < n; i++) {
            Cube qi = formeBase.get(i);
            boolean bTrouvé = false;
            for (int j = 0; j < n; j++) {
                if ( qi.positionIdentique(p.formeBase.get(j)) ) {
                    bTrouvé = true;
                    break;
                }
            }
            if ( bTrouvé == false )
                return false;
        }
        return true;
    }
}
